// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#pragma once

#include <math.h>

#include "column/column.h"
#include "column/column_builder.h"
#include "column/column_viewer.h"
#include "exprs/vectorized/binary_function.h"
#include "exprs/vectorized/decimal_binary_function.h"
#include "exprs/vectorized/function_helper.h"
#include "exprs/vectorized/unary_function.h"
#include "udf/udf.h"
#include "util/string_parser.hpp"

namespace starrocks {

class Expr;

namespace vectorized {

struct RValueCheckZeroImpl;
struct RValueCheckZeroDecimalv2Impl;
struct pmodImpl;
struct pmodFloatImpl;
struct fmodImpl;
struct modImpl;
struct modDecimalv2Impl;
struct negativeImpl;

class MathFunctions {
public:
    /**
     * @param columns: []
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(pi);

    /**
     * @param columns: []
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(e);

    /**
     * @param columns: [DoubleColumn]
     * @return FloatColumn
     */
    DEFINE_VECTORIZED_FN(sign);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_double);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_float);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_largeint);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_bigint);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_int);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_smallint);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_tinyint);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(abs_decimalv2val);

    DEFINE_VECTORIZED_FN(abs_decimal32);
    DEFINE_VECTORIZED_FN(abs_decimal64);
    DEFINE_VECTORIZED_FN(abs_decimal128);
    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(sin);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(asin);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(cos);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(acos);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(tan);

    /**
     * @param columns: [DoubleColumn]
     * @return DoubleColumn
     */
    DEFINE_VECTORIZED_FN(atan);

    /**
    * @param columns: [DoubleColumn]
    * @return BigIntColumn
    */
    DEFINE_VECTORIZED_FN(ceil);

    /**
    * @param columns: [DoubleColumn]
    * @return BigIntColumn
    */
    DEFINE_VECTORIZED_FN(floor);

    /**
    * @param columns: [DoubleColumn]
    * @return BigIntColumn
    */
    DEFINE_VECTORIZED_FN(round);

    /**
    * @param: [DoubleColumn, IntColumn]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(round_up_to);

    /**
     * @param: [DoubleColumn, IntColumn]
     * @return: DoubleColumn
     */
    DEFINE_VECTORIZED_FN(truncate);

    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(ln);

    /**
    * @param: [DoubleColumn base, DoubleColumn value]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(log);
    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    */

    DEFINE_VECTORIZED_FN(log2);
    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    *
    */

    DEFINE_VECTORIZED_FN(log10);
    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(exp);
    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(radians);
    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(degrees);
    /**
    * @param: [DoubleColumn]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(sqrt);
    /**
    * @param: [DoubleColumn base, DoubleColumn exp]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(pow);
    /**
    * @param: [DoubleColumn y, DoubleColumn x]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(atan2);
    /**
    * @param: [DoubleColumn value]
    * @return: DoubleColumn
    */
    DEFINE_VECTORIZED_FN(cot);

    // @todo: these functions belong to math function?
    // =====================================

    // rand's auxiliary method
    static Status rand_prepare(starrocks_udf::FunctionContext* context,
                               starrocks_udf::FunctionContext::FunctionStateScope scope);
    static Status rand_close(starrocks_udf::FunctionContext* context,
                             starrocks_udf::FunctionContext::FunctionStateScope scope);
    /**
     * @param: []
     * @return: DoubleColumn
     * Get the pseudo-random number that normalize to [0,1].
     */
    DEFINE_VECTORIZED_FN(rand);

    /**
     * @param: [Int64Column]
     * @return: DoubleColumn
     * Get the pseudo-random number that normalize to [0,1] with a seed
     */
    DEFINE_VECTORIZED_FN(rand_seed);

    static void generate_randoms(ColumnBuilder<TYPE_DOUBLE>* result, int32_t num_rows, uint32_t* seed) {
        for (int i = 0; i < num_rows; ++i) {
            *seed = ::rand_r(seed);
            // Normalize to [0,1].
            result->append(static_cast<double>(*seed) / RAND_MAX);
        }
    }

    //
    /**
     * @param: [BigIntColumn]
     * @return: StringColumn
     * Binary representation of BIGINT as string.
     */
    DEFINE_VECTORIZED_FN(bin);
    /**
     * @param: [BigIntColumn, TinyInt1Column, TinyInt2Column]
     * @return: StringColumn
     * Get the string based on TinyInt2 from BigInt based on TinyInt1.
     */
    DEFINE_VECTORIZED_FN(conv_int);
    /**
     * @param: [StringColumn, TinyInt1Column, TinyInt2Column]
     * @return: StringColumn
     * Get the string based on TinyInt2 from String based on TinyInt1.
     */
    DEFINE_VECTORIZED_FN(conv_string);
    //    /**
    //     * @param: [BigIntColumn, BigIntColumn]
    //     * @return: BigIntColumn
    //     */

    /**
     * @tparam : TYPE_DOUBLE, TYPE_BIGINT
     * @param: [TypeColumn, TypeColumn]
     * @return: TypeColumn
     */
    template <PrimitiveType Type>
    DEFINE_VECTORIZED_FN(pmod) {
        auto l = VECTORIZED_FN_ARGS(0);
        auto r = VECTORIZED_FN_ARGS(1);

        if constexpr (Type == TYPE_FLOAT || Type == TYPE_DOUBLE) {
            return VectorizedUnstrictBinaryFunction<RValueCheckZeroImpl, pmodFloatImpl>::evaluate<Type>(l, r);
        } else {
            return VectorizedUnstrictBinaryFunction<RValueCheckZeroImpl, pmodImpl>::evaluate<Type>(l, r);
        }
    }

    /**
     * @tparam : TYPE_FLOAT, TYPE_DOUBLE
     * @param: [TypeColumn, TypeColumn]
     * @return: TypeColumn
     */
    template <PrimitiveType Type>
    DEFINE_VECTORIZED_FN(fmod) {
        auto l = VECTORIZED_FN_ARGS(0);
        auto r = VECTORIZED_FN_ARGS(1);

        return VectorizedUnstrictBinaryFunction<RValueCheckZeroImpl, fmodImpl>::evaluate<Type>(l, r);
    }

    /**
     * @tparam : TYPE_TINYINT, TYPE_SMALLINT, TYPE_INT, TYPE_BIGINT,
     *           TYPE_LARGEINT, TYPE_FLOAT, TYPE_DOUBLE
     * @todo: TYPE_DECIMALV2
     * @param: [TypeColumn, TypeColumn]
     * @return: TypeColumn
     */
    template <PrimitiveType Type>
    DEFINE_VECTORIZED_FN(mod) {
        auto l = VECTORIZED_FN_ARGS(0);
        auto r = VECTORIZED_FN_ARGS(1);

        if constexpr (pt_is_decimalv2<Type>) {
            return VectorizedUnstrictBinaryFunction<RValueCheckZeroDecimalv2Impl, modDecimalv2Impl>::evaluate<Type>(l,
                                                                                                                    r);
        } else if constexpr (pt_is_decimal<Type>) {
            // TODO(by satanson):
            //  FunctionContext carry decimal_overflow_check flag to control overflow checking.
            using VectorizedDiv = VectorizedUnstrictDecimalBinaryFunction<Type, ModOp, false>;
            return VectorizedDiv::template evaluate<Type>(l, r);
        } else {
            return VectorizedUnstrictBinaryFunction<RValueCheckZeroImpl, modImpl>::evaluate<Type>(l, r);
        }
    }

    /**
     * @tparam : TYPE_BIGINT, TYPE_DOUBLE
     * @todo: TYPE_DECIMALV2
     * @param: [TypeColumn]
     * @return: TypeColumn
     */
    template <PrimitiveType Type>
    DEFINE_VECTORIZED_FN(positive) {
        return VECTORIZED_FN_ARGS(0);
    }

    /**
     * @tparam : TYPE_BIGINT, TYPE_DOUBLE
     * @todo: TYPE_DECIMALV2
     * @param: [TypeColumn]
     * @return: TypeColumn
     */
    template <PrimitiveType Type>
    DEFINE_VECTORIZED_FN(negative) {
        if constexpr (pt_is_decimal<Type>) {
            const auto& type = context->get_return_type();
            return VectorizedStrictUnaryFunction<negativeImpl>::evaluate<Type>(VECTORIZED_FN_ARGS(0), type.precision,
                                                                               type.scale);
        } else {
            return VectorizedStrictUnaryFunction<negativeImpl>::evaluate<Type>(VECTORIZED_FN_ARGS(0));
        }
    }

    /**
    * @tparam : TYPE_TINYINT, TYPE_SMALLINT, TYPE_INT, TYPE_BIGINT, TYPE_LARGEINT,
    *           TYPE_FLOAT, TYPE_DOUBLE, TYPE_STRING, TYPE_DATETIME, TYPE_DECIMALV2
    * @param: [TypeColumn, ...]
    * @return: TypeColumn
    */
    template <PrimitiveType Type>
    static ColumnPtr least(FunctionContext* context, const Columns& columns) {
        if (columns.size() == 1) {
            return columns[0];
        }

        RETURN_IF_COLUMNS_ONLY_NULL(columns);

        const auto& type = context->get_return_type();
        ColumnBuilder<Type> result(type.precision, type.scale);

        std::vector<ColumnViewer<Type>> list;
        list.reserve(columns.size());
        for (ColumnPtr col : columns) {
            list.emplace_back(ColumnViewer<Type>(col));
        }

        auto size = columns[0]->size();
        for (int row = 0; row < size; row++) {
            auto value = list[0].value(row);
            bool is_null = false;
            for (auto& view : list) {
                is_null = is_null || view.is_null(row);
                // TODO (by satanson):
                //  compare two decimal with different precision and scale and prevent casting into
                //  the same precision and scale, because casting can generate overflow values.
                value = value > view.value(row) ? view.value(row) : value;
            }

            result.append(value, is_null);
        }

        return result.build(ColumnHelper::is_all_const(columns));
    }

    /**
     * @tparam : TYPE_TINYINT, TYPE_SMALLINT, TYPE_INT, TYPE_BIGINT, TYPE_LARGEINT,
     *           TYPE_FLOAT, TYPE_DOUBLE, TYPE_DECIMALV2, TYPE_STRING, TYPE_DATETIME
     * @param: [TypeColumn, ...]
     * @return: TypeColumn
     */
    template <PrimitiveType Type>
    static ColumnPtr greatest(FunctionContext* context, const Columns& columns) {
        if (columns.size() == 1) {
            return columns[0];
        }

        RETURN_IF_COLUMNS_ONLY_NULL(columns);

        const auto& type = context->get_return_type();
        ColumnBuilder<Type> result(type.precision, type.scale);

        std::vector<ColumnViewer<Type>> list;
        list.reserve(columns.size());
        for (ColumnPtr col : columns) {
            list.emplace_back(ColumnViewer<Type>(col));
        }

        auto size = columns[0]->size();
        for (int row = 0; row < size; row++) {
            auto value = list[0].value(row);
            bool is_null = false;
            for (auto& view : list) {
                is_null = is_null || view.is_null(row);
                // TODO (by satanson):
                //  compare two decimal with different precision and scale and prevent casting into
                //  the same precision and scale, because casting can generate overflow values.
                value = value < view.value(row) ? view.value(row) : value;
            }

            result.append(value, is_null);
        }

        return result.build(ColumnHelper::is_all_const(columns));
    }

    static double double_round(double value, int64_t dec, bool dec_unsigned, bool truncate);
    static bool decimal_in_base_to_decimal(int64_t src_num, int8_t src_base, int64_t* result);
    static bool handle_parse_result(int8_t dest_base, int64_t* num, StringParser::ParseResult parse_res);
    static std::string decimal_to_base(int64_t src_num, int8_t dest_base);

private:
    static const int32_t MIN_BASE = 2;
    static const int32_t MAX_BASE = 36;
    static const char* _s_alphanumeric_chars;

public:
    constexpr static double EPSILON = 1e-9;
};

DEFINE_BINARY_FUNCTION_WITH_IMPL(RValueCheckZeroImpl, a, b) {
    return b == 0;
}

DEFINE_BINARY_FUNCTION_WITH_IMPL(RValueCheckZeroDecimalv2Impl, a, b) {
    return b == DecimalV2Value::ZERO;
}

// pmod
DEFINE_BINARY_FUNCTION_WITH_IMPL(pmodImpl, a, b) {
    return ((a % (b + (b == 0))) + b) % (b + (b == 0));
}

DEFINE_BINARY_FUNCTION_WITH_IMPL(pmodFloatImpl, a, b) {
    return ::fmod(::fmod(a, (b + (b == 0))) + b, (b + (b == 0)));
}

// fmod
DEFINE_BINARY_FUNCTION(fmodImpl, fmod);

// mod
DEFINE_BINARY_FUNCTION_WITH_IMPL(modImpl, a, b) {
    return (a % (b + (b == 0)));
}

// mod for DecimalV2Value
DEFINE_BINARY_FUNCTION_WITH_IMPL(modDecimalv2Impl, a, b) {
    return a % (b + ((b == DecimalV2Value::ZERO) ? DecimalV2Value::ONE : DecimalV2Value::ZERO));
}

// negative
DEFINE_UNARY_FN_WITH_IMPL(negativeImpl, v) {
    return -v;
}

} // namespace vectorized
} // namespace starrocks
